import ntp.packet
import threading
from contextlib import AbstractContextManager
import time
import argparse
import os
import sys
import pwd
import grp
import logging
import logging.handlers
import pyagentx3
import ipaddress



BASE_OID_ENTERPRISE             = '1.3.6.1.4.1'
BASE_OID_HOTTIS                 = BASE_OID_ENTERPRISE + '.9676'
BASE_OID_HOTTIS_NTPSEC          = BASE_OID_HOTTIS + '.123'
LOCAL_PREFIX                    = '1'
PEERS_PREFIX                    = '2'


def int_scale1k(x):
    return int (x * 1000)

def int_scale1M(x):
    return int (x * 1000000)

def pass_value(x):
    return x

LOCAL_SERVER_KEYS = [ 
    ['leap', pyagentx3.TYPE_INTEGER, pass_value],
    ['stratum', pyagentx3.TYPE_INTEGER, pass_value],
    ['precision', pyagentx3.TYPE_INTEGER, pass_value],
    ['rootdelay', pyagentx3.TYPE_INTEGER, int_scale1k],
    ['rootdisp', pyagentx3.TYPE_INTEGER, int_scale1k], 
    ['rootdist', pyagentx3.TYPE_INTEGER, int_scale1k],
    ['refid', pyagentx3.TYPE_OCTETSTRING, pass_value],
    ['reftime', pyagentx3.TYPE_OCTETSTRING, pass_value],
    ['tc', pyagentx3.TYPE_INTEGER, pass_value],
    ['peer', pyagentx3.TYPE_INTEGER, pass_value],
    ['offset', pyagentx3.TYPE_INTEGER, int_scale1M],
    ['frequency', pyagentx3.TYPE_INTEGER, int_scale1k],
    ['sys_jitter', pyagentx3.TYPE_INTEGER, int_scale1M],
    ['clk_jitter', pyagentx3.TYPE_INTEGER, int_scale1M],
    ['clock', pyagentx3.TYPE_OCTETSTRING, pass_value],
    ['processor', pyagentx3.TYPE_OCTETSTRING, pass_value],
    ['system', pyagentx3.TYPE_OCTETSTRING, pass_value],
    ['release', pyagentx3.TYPE_OCTETSTRING, pass_value],
    ['version', pyagentx3.TYPE_OCTETSTRING, pass_value],
    ['clk_wander', pyagentx3.TYPE_INTEGER, int_scale1M],
    ['daemon_version', pyagentx3.TYPE_OCTETSTRING, pass_value],
    ['tai', pyagentx3.TYPE_INTEGER, pass_value],
    ['leapsec', pyagentx3.TYPE_OCTETSTRING, pass_value],
    ['expire', pyagentx3.TYPE_OCTETSTRING, pass_value],
    ['mintc', pyagentx3.TYPE_INTEGER, pass_value]
]

PEER_KEYS = [
    ['srcadr', pyagentx3.TYPE_IPADDRESS, ipaddress.ip_address],
    ['srcport', pyagentx3.TYPE_INTEGER, pass_value],
    ['dstadr', pyagentx3.TYPE_IPADDRESS, ipaddress.ip_address],
    ['dstport', pyagentx3.TYPE_INTEGER, pass_value],
    ['leap', pyagentx3.TYPE_INTEGER, pass_value],
    ['hmode', pyagentx3.TYPE_INTEGER, pass_value],
    ['stratum', pyagentx3.TYPE_INTEGER, pass_value],
    ['ppoll', pyagentx3.TYPE_INTEGER, pass_value],
    ['hpoll', pyagentx3.TYPE_INTEGER, pass_value],
    ['precision', pyagentx3.TYPE_INTEGER, pass_value],
    ['rootdelay', pyagentx3.TYPE_INTEGER, int_scale1k],
    ['rootdisp', pyagentx3.TYPE_INTEGER, int_scale1k],
    ['refid', pyagentx3.TYPE_OCTETSTRING, pass_value],
    ['reftime', pyagentx3.TYPE_OCTETSTRING, pass_value],
    ['rec', pyagentx3.TYPE_OCTETSTRING, pass_value],
    ['xmt', pyagentx3.TYPE_OCTETSTRING, pass_value],
    ['reach', pyagentx3.TYPE_INTEGER, pass_value],
    ['unreach', pyagentx3.TYPE_INTEGER, pass_value],
    ['delay-s', pyagentx3.TYPE_OCTETSTRING, pass_value],
    ['delay', pyagentx3.TYPE_INTEGER, int_scale1k],
    ['offset', pyagentx3.TYPE_INTEGER, int_scale1M],
    ['jitter', pyagentx3.TYPE_INTEGER, int_scale1M],
    ['dispersion', pyagentx3.TYPE_INTEGER, int_scale1k],
    ['keyid', pyagentx3.TYPE_INTEGER, pass_value],
    ['filtdelay', pyagentx3.TYPE_OCTETSTRING, pass_value],
    ['filtoffset', pyagentx3.TYPE_OCTETSTRING, pass_value],
    ['pmode', pyagentx3.TYPE_INTEGER, pass_value],
    ['filtdisp', pyagentx3.TYPE_OCTETSTRING, pass_value],
    ['flash', pyagentx3.TYPE_INTEGER, pass_value],
    ['headway', pyagentx3.TYPE_INTEGER, pass_value],
    ['ntscookies', pyagentx3.TYPE_INTEGER, pass_value]
]


class DataStore(AbstractContextManager):
    def __init__(self):
        self.lock = threading.Lock()
        self.data = {}

    def update_data(self, data):
        self.data.clear()
        self.data.update(data)

    def __enter__(self):
        self.lock.acquire()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.lock.release()


class NtpDataCollector(threading.Thread):
    def __init__(self, ntpserver='localhost', period=60):
        self.ntpserver = ntpserver
        self.period = period
        self.stop_event = threading.Event()

        self.session = ntp.packet.ControlSession()
        self.session.openhost(self.ntpserver)

        self.data_store = DataStore()

        threading.Thread.__init__(self)

    def run(self):
        while not self.stop_event.is_set():
            logger.debug('Query ntp server')

            ntpserver_vars = self.session.readvar(0)
            logger.debug(f"{ntpserver_vars=}")

            tmp_data_store = {}
            tmp_data_store['local'] = dict(ntpserver_vars)

            peers = self.session.readstat()
            tmp_data_store['peers'] = {}
            for peer in peers:
                peer_vars = self.session.readvar(peer.associd)
                tmp_data_store['peers'][peer.associd] = dict(peer_vars)
                logger.debug(f"{peer.associd=}, {peer_vars=}")

            with self.data_store as ds:
                ds.update_data(tmp_data_store)

            time.sleep(self.period)
        logger.info('NtpDataCollector terminating')

    def get_data_store(self):
        return self.data_store

    def stop(self):
        self.stop_event.set()
        self.join()

class TestUpdater(pyagentx3.Updater):
    def update(self):
        self.set_INTEGER('1', 123)

class NtpsecDataUpdater(pyagentx3.Updater):
    def update(self):
        with self.data_store as ds:
            if ds.data:
                try:
                    
                    for index, data_spec in enumerate(LOCAL_SERVER_KEYS, start=1):
                        logger.debug(f"local: {index=} {data_spec=}")
                        oid_prefix = f"{LOCAL_PREFIX}.{index}"
                        self._data[oid_prefix] = {
                            'name': oid_prefix, 
                            'type': data_spec[1], 
                            'value': data_spec[2](ds.data['local'][data_spec[0]])
                        }
                    for peer_index, (associd, peer) in enumerate(ds.data['peers'].items(), start=1):
                        logger.debug(f"peer: {peer}")
                        index_oid_prefix = f"{PEERS_PREFIX}.1.{peer_index}"
                        self._data[index_oid_prefix] = {
                            'name': index_oid_prefix, 
                            'type': pyagentx3.TYPE_INTEGER,
                            'value': associd
                        }
                        for key_index, data_spec in enumerate(PEER_KEYS, start=2):
                            logger.debug(f"peer: {associd=} {key_index=} {data_spec=}")
                            oid_prefix = f"{PEERS_PREFIX}.{key_index}.{peer_index}"
                            self._data[oid_prefix] = {
                                'name': oid_prefix, 
                                'type': data_spec[1], 
                                'value': data_spec[2](peer[data_spec[0]])
                            }
                except Exception as e:
                    logger.error(f"Failed to update: {type(e)} {e}")



class NtpsecAgent(pyagentx3.Agent):
    def __init__(self, agent_id='NtpsecAgent', socket_path=None, data_store=None):
        logger.debug('Agent created')
        self.data_store = data_store
        super().__init__(agent_id, socket_path)

    def setup(self):
        logger.debug('Agent setup')
        self.register(BASE_OID_HOTTIS_NTPSEC, NtpsecDataUpdater, freq=1, data_store=self.data_store)




def daemonize(pid_filename):
    if os.fork() > 0:
        sys.exit(0)

    os.setsid()
    pid = os.fork()
    if pid > 0:
        with open(pid_filename, 'w') as pid_file:
            pid_file.write(str(pid))
        sys.exit(0)

    sys.stdout.flush()
    sys.stderr.flush()

    with open("/dev/null", "r") as devnull:
        os.dup2(devnull.fileno(), sys.stdin.fileno())
    with open("/tmp/agentx-ntpsec.log", "a+") as log:
        os.dup2(log.fileno(), sys.stdout.fileno())
        os.dup2(log.fileno(), sys.stderr.fileno())

    logger.removeHandler(stdout_handler)
    pyagentx3.setup_logging(debug=True)

def set_user_group(user, group):
    if group:
        try:
            gid = grp.getgrnam(group).gr_gid
        except KeyError:
            logger.error(f"Group {group} does not exist")
            sys.exit(1)
        os.setgid(gid)
    if user:
        try:
            uid = pwd.getpwnam(user).pw_uid
        except KeyError:
            logger.error(f"user {user} does not exist")
            sys.exit(1)
        os.setuid(uid)


if __name__ == '__main__':
    logging.basicConfig(
            level=logging.DEBUG, 
            format="%(name)s - %(levelname)s - %(message)s",
            handlers=[logging.handlers.SysLogHandler(address='/dev/log')]
    )
    logger = logging.getLogger('agentx-ntpsec')
    stdout_handler = logging.StreamHandler(sys.stdout)
    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    stdout_handler.setFormatter(formatter)
    logger.addHandler(stdout_handler)

    pid_filename = '/tmp/agentx-ntpsec.pid'

    parser = argparse.ArgumentParser(description='snmpd agentx extension for ntpsec')
    parser.add_argument('--period', '-p',
                        help='Period to query the NTP server, in seconds, default 60s',
                        required=False,
                        default=60)
    parser.add_argument('--ntpserver', '-n',
                        help='NTP server to query, default is localhost',
                        required=False,
                        default='localhost')
    parser.add_argument('--daemonize', '-d',
                        help='Run process in background',
                        required=False,
                        action='store_true',
                        default=False)
    parser.add_argument('--pid',
                        help=f"pid-file when running as daemon, default is {pid_filename}",
                        required=False,
                        default=pid_filename)
    parser.add_argument('-u', '--user',
                        help="Set uid of process",
                        required=False,
                        default='')
    parser.add_argument('-g', '--group',
                        help="Set gid of process",
                        required=False,
                        default='')

    args = parser.parse_args()
    
    if args.daemonize:
        daemonize(pid_filename)
    set_user_group(args.user, args.group)

    if args.group:
        try:
            gid = grp.getgrnam(args.group).gr_gid
        except KeyError:
            logger.error(f"Group {args.group} does not exist")
            sys.exit(1)
        os.setgid(gid)
    if args.user:
        try:
            uid = pwd.getpwnam(args.user).pw_uid
        except KeyError:
            logger.error(f"user {args.user} does not exist")
            sys.exit(1)
        os.setuid(uid)
        

    ntpserver = args.ntpserver
    period = args.period
    

    try:
        ndc = NtpDataCollector(ntpserver=ntpserver, period=period)
        ndc.start()

        nsax = NtpsecAgent(data_store=ndc.get_data_store())
        nsax.start()
    except Exception as e:
        logger.error(f"Unhandled exception: {e}")
        nsax.stop()
        ndc.stop()
    except KeyboardInterrupt:
        nsax.stop()
        ndc.stop()

